from typing import Annotated, Any

from astropy.units import Quantity, Unit
from pydantic import GetCoreSchemaHandler
from pydantic_core import core_schema


class _AstropyQuantityPydanticTypeAnnotation(type):
    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        _source_type: Any,
        _handler: GetCoreSchemaHandler,
    ) -> core_schema.CoreSchema:
        dict_schema = core_schema.chain_schema(
            [
                core_schema.typed_dict_schema(
                    {
                        "value": core_schema.typed_dict_field(
                            core_schema.union_schema(
                                [
                                    core_schema.float_schema(),
                                    core_schema.list_schema(),
                                ]
                            )
                        ),
                        "unit": core_schema.typed_dict_field(core_schema.str_schema()),
                    }
                ),
                core_schema.no_info_plain_validator_function(
                    lambda d: Quantity(d["value"], unit=d["unit"])
                ),
            ]
        )

        def validate_from_string(value: str):
            return Quantity(value)

        def json_serialize_value(q: Quantity):
            # Serialize to the native types when going to JSON, numpy
            # arrays are not supported.
            # Must import every time in case someone has changed it!
            from astropydantic import UNIT_STRING_FORMAT

            return {
                "value": q.value.tolist(),
                "unit": q.unit.to_string(format=UNIT_STRING_FORMAT),
            }

        str_schema = core_schema.chain_schema(
            [
                core_schema.str_schema(),
                core_schema.no_info_plain_validator_function(validate_from_string),
            ]
        )

        python_schema = core_schema.union_schema(
            [
                core_schema.is_instance_schema(Quantity),
                dict_schema,
                str_schema,
            ]
        )
        json_schema = core_schema.union_schema([dict_schema, str_schema])

        return core_schema.json_or_python_schema(
            json_schema=json_schema,
            python_schema=python_schema,
            serialization=core_schema.plain_serializer_function_ser_schema(
                json_serialize_value, when_used="json-unless-none"
            ),
        )

    def __getitem__(cls, unit: Unit):
        """Implement support for AstroPydanticQuantity[<unit>]."""

        class _AstropyQuantityPydanticTypeAnnotationWithUnit:
            @classmethod
            def __get_pydantic_core_schema__(cls, _source_type: Any, _handler: Any):
                base = (
                    _AstropyQuantityPydanticTypeAnnotation.__get_pydantic_core_schema__(
                        _source_type, _handler
                    )
                )
                with_unit = core_schema.chain_schema(
                    [
                        base,
                        core_schema.no_info_plain_validator_function(
                            lambda q: q.to(unit)
                        ),
                    ]
                )
                with_unit["serialization"] = base["serialization"]
                return with_unit

        return Annotated[Quantity, _AstropyQuantityPydanticTypeAnnotationWithUnit]


class AstroPydanticQuantity(Quantity, metaclass=_AstropyQuantityPydanticTypeAnnotation):
    pass



"""
Test that (de)serialization works
"""

from astropy import units as u
from pydantic import BaseModel

from astropydantic import AstroPydanticUnit


def test_unit_string():
    class TestModel(BaseModel):
        x: AstroPydanticUnit

    m = TestModel(x="m")

    assert isinstance(m.x, u.core.UnitBase)

    serialized = m.model_dump()
    TestModel.model_validate(serialized)

    json = m.model_dump_json()
    TestModel.model_validate_json(json)


def test_unit_type():
    class TestModel(BaseModel):
        x: AstroPydanticUnit

    m = TestModel(x=u.m)

    assert isinstance(m.x, u.core.UnitBase)

    serialized = m.model_dump()

    reconstructed = TestModel.model_validate(serialized)

    assert isinstance(reconstructed.x, u.core.UnitBase)

    json = m.model_dump_json()
    TestModel.model_validate_json(json)


def test_unit_string_format():
    class TestModel(BaseModel):
        x: AstroPydanticUnit

    m = TestModel(x=u.km / u.s**2)

    assert m.model_dump_json() == '{"x":"km.s**-2"}'

    import astropydantic

    astropydantic.UNIT_STRING_FORMAT = "fits"

    class TestModel(BaseModel):
        x: AstroPydanticUnit

    m = TestModel(x=u.km / u.s**2)

    assert m.model_dump_json() != '{"x":"km.s**-2"}'



from typing import Any

from astropy.units import Unit, UnitBase
from pydantic import GetCoreSchemaHandler
from pydantic_core import core_schema


class _AstropyUnitPydanticTypeAnnotation(type):
    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        _source_type: Any,
        _handler: GetCoreSchemaHandler,
    ) -> core_schema.CoreSchema:
        def validate_from_string(value: str) -> Unit:
            return Unit(s=value)

        from_string_schema = core_schema.chain_schema(
            [
                core_schema.str_schema(),
                core_schema.no_info_plain_validator_function(validate_from_string),
            ]
        )

        def serialize_to_string(value: Unit) -> str:
            # Must import every time in case someone has changed it!
            from astropydantic import UNIT_STRING_FORMAT

            return value.to_string(format=UNIT_STRING_FORMAT)

        return core_schema.json_or_python_schema(
            json_schema=from_string_schema,
            python_schema=core_schema.union_schema(
                [
                    core_schema.is_instance_schema(UnitBase),
                    from_string_schema,
                ]
            ),
            # If we're just serializing to a dict, keep it as a Unit.
            serialization=core_schema.plain_serializer_function_ser_schema(
                serialize_to_string, when_used="json-unless-none"
            ),
        )


class AstroPydanticUnit(UnitBase, metaclass=_AstropyUnitPydanticTypeAnnotation):
    pass



from typing import Any

from astropy.coordinates import ICRS, SkyCoord
from pydantic import GetCoreSchemaHandler
from pydantic_core import core_schema

from .quantity import AstroPydanticQuantity


class _AstropyICRSPydanticTypeAnnotation(type(ICRS)):
    @classmethod
    def __get_pydantic_core_schema__(
        cls, _source_type: Any, _handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        quantity_schema = _handler.generate_schema(AstroPydanticQuantity)

        dict_schema = core_schema.chain_schema(
            [
                core_schema.typed_dict_schema(
                    {
                        "ra": core_schema.typed_dict_field(quantity_schema),
                        "dec": core_schema.typed_dict_field(quantity_schema),
                    }
                ),
                core_schema.no_info_plain_validator_function(
                    lambda x: ICRS(ra=x["ra"], dec=x["dec"])
                ),
            ]
        )

        def json_serialize_value(c: ICRS):
            # Serialize to the native types whne going to JSON.
            from astropydantic import UNIT_STRING_FORMAT

            return {
                "ra": {
                    "value": c.ra.value,
                    "unit": c.ra.unit.to_string(format=UNIT_STRING_FORMAT),
                },
                "dec": {
                    "value": c.dec.value,
                    "unit": c.dec.unit.to_string(format=UNIT_STRING_FORMAT),
                },
            }

        def validate_icrs_like(value):
            if isinstance(value, ICRS):
                return value
            if isinstance(value, SkyCoord):
                if getattr(value.frame, "name", None) == "icrs":
                    # Coerce to an actual ICRS instance to unify type
                    return ICRS(ra=value.ra, dec=value.dec)
                raise TypeError("SkyCoord must have frame='icrs'")
            raise TypeError("Expected ICRS or SkyCoord(frame='icrs')")

        python_schema = core_schema.union_schema(
            [
                core_schema.no_info_plain_validator_function(validate_icrs_like),
                dict_schema,
            ]
        )

        return core_schema.json_or_python_schema(
            json_schema=dict_schema,
            python_schema=python_schema,
            serialization=core_schema.plain_serializer_function_ser_schema(
                json_serialize_value, when_used="json-unless-none"
            ),
        )


class AstroPydanticICRS(ICRS, metaclass=_AstropyICRSPydanticTypeAnnotation):
    pass



from typing import Annotated, Any

from astropy.time import Time, TimeBase
from pydantic import GetCoreSchemaHandler
from pydantic_core import CoreSchema, core_schema


class _AstropyTimePydanticTypeAnnotation:
    @classmethod
    def __get_pydantic_core_schema__(
        cls, _source_type: Any, _handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        from_string = core_schema.chain_schema(
            [
                core_schema.str_schema(),
                core_schema.no_info_plain_validator_function(lambda s: Time(s)),
            ]
        )

        from_datetime = core_schema.chain_schema(
            [
                core_schema.datetime_schema(),
                core_schema.no_info_plain_validator_function(lambda d: Time(d)),
            ]
        )

        def serialize(v: Time):
            from astropydantic import TIME_OUTPUT_FORMAT, TIME_OUTPUT_SCALE

            fmt = TIME_OUTPUT_FORMAT
            try:
                v = getattr(v, TIME_OUTPUT_SCALE)
            except AttributeError:
                raise ValueError(f"Unsupported TIME_OUTPUT_SCALE: {TIME_OUTPUT_SCALE}")

            # handle special case of precision with isot
            if "iso" in fmt:
                fmt, _, precision = fmt.partition("_")
                precision = int(precision)
                v = Time(v, precision=precision, copy=False)

            try:
                return getattr(v, fmt)
            except AttributeError:
                raise ValueError(
                    f"Unsupported TIME_OUTPUT_FORMAT: {TIME_OUTPUT_FORMAT}"
                )

        return core_schema.json_or_python_schema(
            python_schema=core_schema.union_schema(
                [core_schema.is_instance_schema(TimeBase), from_string, from_datetime]
            ),
            json_schema=core_schema.union_schema([from_string, from_datetime]),
            serialization=core_schema.plain_serializer_function_ser_schema(
                serialize,
            ),
        )


AstroPydanticTime = Annotated[Time, _AstropyTimePydanticTypeAnnotation]



"""
Astropy time integration
"""

import datetime

import pytest
from astropy import time as t
from pydantic import BaseModel

from astropydantic import AstroPydanticTime


def test_unit_string():
    class TestModel(BaseModel):
        x: AstroPydanticTime

    m = TestModel(x=t.Time.now())

    assert isinstance(m.x, t.TimeBase)

    serialized = m.model_dump()
    TestModel.model_validate(serialized)

    json = m.model_dump_json()
    TestModel.model_validate_json(json)


def test_unit_datetime(monkeypatch):
    import astropydantic

    monkeypatch.setattr(astropydantic, "TIME_OUTPUT_FORMAT", "datetime")

    class TestModel(BaseModel):
        x: AstroPydanticTime

    m = TestModel(x=datetime.datetime.now())

    assert isinstance(m.x, t.TimeBase)

    serialized = m.model_dump()
    TestModel.model_validate(serialized)

    assert isinstance(serialized["x"], datetime.datetime)

    json = m.model_dump_json()
    TestModel.model_validate_json(json)

    # since TIME_OUTPUT_FORMAT is datetime, also an mjd format datetime should
    # be converted to datetime
    m = TestModel(x=t.Time(58000, format="mjd"))
    serialized = m.model_dump()
    assert isinstance(serialized["x"], datetime.datetime)


def test_unit_mjd(monkeypatch):
    import astropydantic

    monkeypatch.setattr(astropydantic, "TIME_OUTPUT_FORMAT", "mjd")

    class TestModel(BaseModel):
        t: AstroPydanticTime

    now = t.Time.now()
    m = TestModel(t=now)
    serialized = m.model_dump()
    assert serialized["t"] == now.utc.mjd


def test_scale(monkeypatch):
    import astropydantic

    monkeypatch.setattr(astropydantic, "TIME_OUTPUT_SCALE", "tai")

    class TestModel(BaseModel):
        t: AstroPydanticTime

    now = t.Time.now()
    m = TestModel(t=now.utc)
    serialized = m.model_dump()

    now.precision = 9
    assert serialized["t"] == now.tai.isot


def test_unit_disallowed():
    import astropydantic

    astropydantic.TIME_OUTPUT_FORMAT = "asdfasdgasdgas"

    class TestModel(BaseModel):
        x: AstroPydanticTime

    m = TestModel(x=datetime.datetime.now())

    assert isinstance(m.x, t.TimeBase)

    with pytest.raises(ValueError):
        m.model_dump()



"""
Test that (de)serialization works
"""

import pytest
from astropy import units as u
from pydantic import BaseModel, ValidationError

from astropydantic import AstroPydanticQuantity


def test_quantity_core():
    class TestModel(BaseModel):
        x: AstroPydanticQuantity

    m = TestModel(x={"value": 1.0, "unit": "m"})

    assert isinstance(m.x, u.Quantity)

    serialized = m.model_dump()
    TestModel.model_validate(serialized)

    json = m.model_dump_json()
    TestModel.model_validate_json(json)


def test_quantity_from_string():
    class TestModel(BaseModel):
        x: AstroPydanticQuantity

    m = TestModel(x="1.0 m")

    assert isinstance(m.x, u.Quantity)

    serialized = m.model_dump()
    TestModel.model_validate(serialized)

    json = m.model_dump_json()
    TestModel.model_validate_json(json)


def test_quantity_obj():
    class TestModel(BaseModel):
        x: AstroPydanticQuantity

    m = TestModel(x=u.Quantity(value=0.1, unit="km"))

    assert isinstance(m.x, u.Quantity)

    serialized = m.model_dump()
    TestModel.model_validate(serialized)

    json = m.model_dump_json()
    TestModel.model_validate_json(json)


def test_quantity_arr():
    class TestModel(BaseModel):
        x: AstroPydanticQuantity

    m = TestModel(x=u.Quantity(value=[0.1, 0.2, 0.3], unit="km"))

    assert isinstance(m.x, u.Quantity)

    serialized = m.model_dump()
    TestModel.model_validate(serialized)

    json = m.model_dump_json()
    TestModel.model_validate_json(json)


def test_quantity_ndarr():
    class TestModel(BaseModel):
        x: AstroPydanticQuantity

    m = TestModel(x=u.Quantity(value=[[0.1, 0.2, 0.3], [0.1, 0.2, 0.3]], unit="km"))

    assert isinstance(m.x, u.Quantity)

    serialized = m.model_dump()
    TestModel.model_validate(serialized)

    json = m.model_dump_json()
    TestModel.model_validate_json(json)


def test_quantity_arr_from_dict():
    class TestModel(BaseModel):
        x: AstroPydanticQuantity

    m = TestModel(x={"value": [0.1, 0.1, 0.2], "unit": "A"})

    assert isinstance(m.x, u.Quantity)

    serialized = m.model_dump()
    TestModel.model_validate(serialized)

    json = m.model_dump_json()
    TestModel.model_validate_json(json)


def test_indexed_units():
    class TestModel(BaseModel):
        x: AstroPydanticQuantity[u.km]

    m = TestModel(x={"value": [0.1, 0.1, 0.2], "unit": "m"})

    assert isinstance(m.x, u.Quantity)

    serialized = m.model_dump()
    TestModel.model_validate(serialized)

    json = m.model_dump_json()
    TestModel.model_validate_json(json)


def test_fail_compare_units():
    class TestModel(BaseModel):
        x: AstroPydanticQuantity[u.g]

    with pytest.raises(ValidationError):
        TestModel(x={"value": [0.1, 0.1, 0.2], "unit": "m"})



UNIT_STRING_FORMAT = "vounit"  # Must be at the top to prevent circular imports
TIME_OUTPUT_FORMAT = "isot_9"  # Must be at the top to prevent circular imports
TIME_OUTPUT_SCALE = "utc"

from .quantity import AstroPydanticQuantity  # noqa: E402 I001
from .unit import AstroPydanticUnit  # noqa: E402 I001
from .time import AstroPydanticTime  # noqa: E402 I001
from .icrs import AstroPydanticICRS  # noqa: E402 I001


__all__ = [
    "AstroPydanticUnit",
    "AstroPydanticQuantity",
    "AstroPydanticTime",
    "AstroPydanticICRS",
    "UNIT_STRING_FORMAT",
]



"""
Tests for ICRS serialization
"""

import pytest
from astropy import units as u
from astropy.coordinates import ICRS, SkyCoord
from pydantic import BaseModel

from astropydantic.icrs import AstroPydanticICRS


def test_icrs_serialization():
    class TestModel(BaseModel):
        x: AstroPydanticICRS

    m = TestModel(x=ICRS(ra=20.0 * u.deg, dec=10.0 * u.deg))

    serialized = m.model_dump()
    TestModel.model_validate(serialized)

    json = m.model_dump_json()
    TestModel.model_validate_json(json)

    reconstituted = TestModel.model_validate_json(json)

    assert reconstituted.x.ra == m.x.ra
    assert reconstituted.x.dec == m.x.dec


def test_skycoord_serialization():
    class TestModel(BaseModel):
        x: AstroPydanticICRS

    coordinate = SkyCoord(ra=20.0 * u.deg, dec=10.0 * u.deg, frame="icrs")

    m = TestModel(x=coordinate)

    serialized = m.model_dump()
    TestModel.model_validate(serialized)

    json = m.model_dump_json()
    TestModel.model_validate_json(json)

    reconstituted = TestModel.model_validate_json(json)

    assert reconstituted.x.ra == m.x.ra
    assert reconstituted.x.dec == m.x.dec


def test_skycoord_failure():
    class TestModel(BaseModel):
        x: AstroPydanticICRS

    coordinate = SkyCoord(ra=20.0 * u.deg, dec=10.0 * u.deg, frame="fk5")

    with pytest.raises(TypeError):
        TestModel(x=coordinate)
